/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"

#include <ctype.h>

#include "dict.h"
#include "str.h"
#include "stack.h"
#include "sysdict.h"

typedef struct dictstack {
	Dict *dict;		/* Array of dictionary entries */
	int n;			/* Number of elements in dictionary */
	int sz;			/* Maximum size of dictionary */
	struct dictstack *next;
	} DictStack;

DictStack sysds= { sysdict, SYSDICTLEN, SYSDICTLEN, NULL };

DictStack *dictstack= &sysds;

Dict *find_dict(char *label, DictStack *s);
Dict *find_dictstack(char *label);
Dict *add_dict(DictStack *s, char *label, Token *t, int flag);

#ifndef isascii
#define isascii(c)      ((unsigned)(c)<=0177)
#endif

#ifdef CHECKDICT
void check_dictstack(void);
#else
#define check_dictstack()
#endif

/* DUMP_DICT - printout all variables in a dictionary.  Debugging only.
 */

void dump_dict(Dict *dict, int size, FILE *fp)
{
int col= 0;
int i,len,typestring= FALSE;
char bf[BFSZ+1];

    for (i= 0; i < size; i++)
    {
	if (class(dict[i].t) == TKC_FUNCTION) continue;

	if (type(dict[i].t) == TK_INTEGER)
	    sprintf(bf,"%s=%d",dict[i].label, ival(dict[i].t));
	else if (type(dict[i].t) == TK_TIME)
	    sprintf(bf,"%s=@%ld",dict[i].label, tval(dict[i].t));
	else if (type(dict[i].t) == TK_STRING)
	{
	    sprintf(bf,"%s=(",dict[i].label);
	    typestring= TRUE;
	}
	else if (type(dict[i].t) == TK_UNBOUND_LITERAL)
	    sprintf(bf,"%s=/%s",dict[i].label, sval(dict[i].t));
	else if (type(dict[i].t) == TK_UNBOUND_SYMBOL)
	    sprintf(bf,"%s=%s",dict[i].label, sval(dict[i].t));
	else if (type(dict[i].t) == TK_MARK)
	    sprintf(bf,"%s=mark",dict[i].label);
	else if (type(dict[i].t) == TK_PROCEDURE)
	    sprintf(bf,"%s={...}",dict[i].label);
	else if (type(dict[i].t) == TK_ARRAY)
	    sprintf(bf,"%s=[...]",dict[i].label);
	else if (type(dict[i].t) == TK_REGEX)
	    sprintf(bf,"%s=<<...>>",dict[i].label);
	else
	    sprintf(bf,"%s=<weird type>",dict[i].label);
	len= strlen(bf);
	if (col+len+1 > 78)
	    { col= 0; putc('\n',fp);}
	else if (col > 0)
	    { col++; putc(' ',fp);}
	fputs(bf,fp);
	col+= len;
	if (typestring)
	{
	    char *cp, *ch, clen;
	    for (cp= sval(dict[i].t); *cp != '\0'; cp++)
	    {
		switch (*cp)
		{
		case '\n': ch= "\\n";  clen= 2; break;
		case '\b': ch= "\\b";  clen= 2; break;
		case '\t': ch= "\\t";  clen= 2; break;
		case '\\': ch= "\\\\"; clen= 2; break;
		case '(':  ch= "\\(";  clen= 2; break;
		case ')':  ch= "\\(";  clen= 2; break;
		case '<':  ch= ttymode ? "<" : "&lt;";  clen= 1; break;
		case '>':  ch= ttymode ? ">" : "&gt;";  clen= 1; break;
		case '&':  ch= ttymode ? "&" : "&amp;"; clen= 1; break;
		default:   ch= NULL; clen= 1; break;
		}
		if (col+clen > 78)
		    {col= 0; fputs("\\\n",fp);}
		if (ch == NULL)
		    putc(*cp, fp);
		else
		    fputs(ch, fp);
		col+= clen;
	    }
	    putc(')',fp);
	    col++;
	    typestring= FALSE;
	}
    }
    if (col != 0)
	putc('\n',fp);
}


/* DUMP_DICT_STACK - Print all variables in all dictionaries on the stack.
 * Debugging only.
 */

void dump_dict_stack(FILE *fp)
{
DictStack *ds;
int n;

    /* Count number of user dictionaries */
    for (ds= dictstack, n= 0; ds != &sysds && ds != NULL; ds= ds->next, n++)
	;

    for (ds= dictstack; n > 0; n--, ds= ds->next)
    {
	fprintf(fp,"User Dictionary %d:\n", n-1);
	dump_dict(ds->dict, ds->n, fp);
	fputc('\n', fp);
    }

    fprintf(fp,"System Dictionary:\n", n);
    dump_dict(sysdict, SYSDICTLEN, fp);
}


/* FIND_DICT - return a pointer to the Dict structure for the named symbol in
 * the given dictionary.  Returns NULL if it is not in that dictionary.
 */

Dict *find_dict(char *label, DictStack *s)
{
int hi,lo,i,sc;

    lo= 0;
    hi= s->n-1;
    while (lo <= hi)
    {
	i= (hi+lo)/2;
	if ((sc= strcmp(label, s->dict[i].label)) == 0)
	    return &(s->dict[i]);
	else if (sc < 0)
	    hi= i-1;
	else
	    lo= i+1;
    }
    return NULL;
}

/* FIND_DICTSTACK - return a pointer to the Dict structure for the named
 * symbol in the topmost dictionary in which it is defined.  Returns NULL
 * if it is not defined.
 */

Dict *find_dictstack(char *label)
{
Dict *d= NULL;
DictStack *s;

    for (s= dictstack; s != NULL; s= s->next)
    {
	if ((d= find_dict(label,s)) != NULL)
	    return d;
    }
    return NULL;
}


/* GET_DICT - Return a pointer to the Token structure for the given symbol
 * name.  This searches the entire dictionary stack, and if it isn't in the
 * dictionary it looks in the environment.  If it isn't there either, it
 * returns NULL.
 */

Token *get_dict(char *label)
{
Dict *d;
static Token t;
#ifdef DFLT_HTTP_HOST
static char dflt_http_host[]= DFLT_HTTP_HOST;
#endif
#ifdef DFLT_SCRIPT_NAME
static char dflt_script_name[]= DFLT_SCRIPT_NAME;
#endif

    /* Search the Dictionary Stack */
    if ((d= find_dictstack(label)) != NULL)
	return &(d->t);

    /* If it isn't there, look in Environment */
    if ((t.val= (void *)getenv(label)) == NULL)
    {
	/* A couple environment variables may have default values */
#ifdef DFLT_HTTP_HOST
	if (!strcmp(label,"HTTP_HOST"))
	    t.val= dflt_http_host;
	else
#endif
#ifdef DFLT_SCRIPT_NAME
	if (!strcmp(label,"SCRIPT_NAME"))
	    t.val= dflt_script_name;
	else
#endif
	    return NULL;
    }
    
    /* Return the environment variable */
    t.flag= TK_STRING;
    return &t;
}


/* GET_SYMBOL - Return a pointer to the Token structure for the given symbol,
 * which may be either a bound or unbound literal or token structure.  Returns
 * NULL if the symbol is unbound and undefined.
 */

Token *get_symbol(Token *s)
{
    if (class(*s) == TKC_INTEGER)
	/* Bound symbol or literal - return sysdict entry */
	return &sysdict[ival(*s)].t;
    else
	return get_dict(sval(*s));
}


/* LABEL_OK - Check if the given character string obeys the rules to be a valid
 * symbol name.
 */

void label_ok(char *l)
{
char *p;

    if (!isascii(l[0]) || !isprint(l[0]) || isdigit(l[0]))
	die("Symbol %s has bad initial character\n",l);
    
    if (ispunct(l[0]))
    {
	if (l[1] != '\0')
	    die("Illegal symbol %s",l);
    }
    else
    {
	for (p= l+1; *p != '\0'; p++)
	    if (!isalnum(*p) && *p != '_' && *p != '.')
		die("Bad character in symbol %s",l);
	    
	if (p - l > MAXSYMLEN)
	    die("Symbol name %s too long",l);
    }
}


/* FUNC_BEGIN:  <dict-size> begin -
 * create a new dictionary on the stack
 */

void func_begin()
{
int size;
DictStack *s;

    size= pop_integer();

    if (size < 0)
	die("Negative dictionary size");
    
    s= (DictStack *)malloc(sizeof(DictStack));

    if (size > 0)
	s->dict= (Dict *)malloc(size*sizeof(Dict));
    else
	s->dict= NULL;
    s->sz= size;
    s->n= 0;
    s->next= dictstack;
    dictstack= s;
}


/* FUNC_END:  - end -
 * pop a dictionary off the stack
 */

void func_end()
{
DictStack *s;
int i;

    if (dictstack == &sysds)
	die("Cannot end system dictionary");
    
    /* Remove dictionary from dictionary stack */
    s= dictstack;
    dictstack= s->next;

    if (s->dict != NULL)
    {
	/* Free values of stuff in dictionary */
	for (i= 0; i < s->n; i++)
	    free_val(&(s->dict[i].t));

	/* Free dictionary array */
	free(s->dict);
    }

    /* Free dictionary stack entry */
    free(s);
}

#ifdef CLEANUP

/* FREE_DICTSTACK - This deallocates everything still on the dictionary stack.
 * We only bother doing this if we are testing for memory leaks.
 */
void free_dictstack()
{
int i;

    /* deallocate all user dictionaries */
    while (dictstack != &sysds)
	func_end();

    /* deallocate all malloced variables in the system dictionary */
    for (i= 0; i < SYSDICTLEN; i++)
	free_val(&(sysdict[i].t));
}

#endif


/* DUP_DICT - If the given system dictionary value was not already in
 * dynamic memory, then copy it into dynamic memory.  Returns a pointer
 * to the value.
 */

char *dup_dict(int i)
{
Token *tk= &(sysdict[i].t);

    if (!(tk->flag & (TKF_FREE)))
    {
	tk->val= strdup(tk->val);
	tk->flag|= TKF_FREE;
    }
    return (char *)tk->val;
}


/* SETS - Set the given system dictionary entry to string value.  This assumes
 * the variable was already a string type.  There is a seti() too, but it's
 * a macro.
 */

char *sets(int i, char *val, int copy)
{
Token *tk= &(sysdict[i].t);

    if ((tk->flag & (TKF_FREE)))
    {
	free(tk->val);
	if (copy == SM_KEEP) tk->flag&= ~TKF_FREE;
    }
    else if (copy != SM_KEEP)
	tk->flag|= TKF_FREE;

    tk->val= (void *)((copy == SM_COPY) ? strdup(val) : val);
    return (char *)(tk->val);
}


/* SET_DICT - Change the given dictionary entry to this token.
 */

void set_dict(Dict *d, Token *t, int flag)
{
    /* May only change these when loading initializing file */
    if ((d->flag & (DT_PROTECT | DT_CONSTANT)) && !initializing)
	die("Cannot change protected variable %s",d->label);

    /* Deallocate memory for old string or array values */
    free_val(&(d->t));

    d->t= *t;
    d->flag|= flag;
}


/* ADD_DICT - insert a new symbol into a dictionary, resizing it if necessary.
 * Return a pointer to the new entry.
 */

Dict *add_dict(DictStack *s, char *label, Token *t, int flag)
{
int i;
    if (s == &sysds)
	die("cannot add symbols to system dictionary");

    s->n++;
    if (s->n > s->sz)
    {
	/* Resize dictionary */
	if (s->sz == 0)
	    s->dict= (Dict *)malloc(s->n*sizeof(Dict));
	else
	    s->dict= (Dict *)realloc(s->dict,s->n*sizeof(Dict));
	s->sz= s->n;
    }
    
    for (i= s->n - 2; i >= 0 && strcmp(label, s->dict[i].label) < 0; i--)
	s->dict[i+1]= s->dict[i];
    i++;
    strcpy(s->dict[i].label,label);
    s->dict[i].flag= flag;
    s->dict[i].t= *t;
    check_dictstack();
    return &(s->dict[i]);
}


/* DEL_DICT - Remove the named symbol from the given dictionary.  Return
 * true if we succeeded.
 */

int del_dict(DictStack *s, char *label)
{
int i,sz;
Dict *d;

    if (s == &sysds)
	die("cannot undefine system dictionary symbol %s",label);

    if ((d= find_dict(label, s)) == NULL)
	return 0;

    if (d->flag & DT_CONSTANT)
	die("cannot undefine constant variable %s", label);

    free_val(&(d->t));

    if ((sz= s->n - (d - s->dict) - 1) > 0)
	memmove(d, d+1, sz*sizeof(Dict));
    s->n--;
    check_dictstack();
    return 1;
}


/* SET_SYS_DICT - sets label to the given value in the system dictionary.
 * value is always given as a string, but will be converted to an integer,
 * if appropriate.  If copy is true and the variable is of a string type,
 * new memory will be allocated for the variable and copied in.
 *
 * If the variable is flagged multi-valued, the new value is appended to the
 * old value, with a comma separator, instead of replacing the old value.
 */

void set_sys_dict(char *label, char *val, int copy)
{
Dict *d;

    /* Find the variable */
    if ((d= find_dict(label,&sysds)) != NULL)
    {
	if ((d->flag & (DT_PROTECT | DT_CONSTANT)) && !initializing)
	    die("Attempt to redefine protected variable %s",label);
	if (class(d->t) == TKC_FUNCTION)
	    die("Attempt to redefine built-in function %s",label);
    }
    else if (dictstack == &sysds)
	die("Cannot save query variable %s -- no user dictionary", label);
    else if (dictstack->n >= dictstack->sz)
	die("Cannot save query variable %s -- user dictionary full",label);
    else if ((d= find_dict(label, dictstack)) == NULL)
    {
	/* Create a new variable */
	Token tk;
	tk.flag= TK_STRING;
	tk.val= (void *)"";
	d= add_dict(dictstack, label, &tk, DT_MULTI);
    }

    /* Variable exists - store into it */
    if (class(d->t) == TKC_INTEGER)
    {
	/* Store value in existing integer variable */
	/* Maybe check integer syntax someday */
	d->t.val= (void *)atoi(val);
    }
    else if (class(d->t) == TKC_TIME)
    {
	/* Store value in existing time variable */
	/* Really need an atotm() function */
	d->t.val= (void *)atol(val);
    }
    else if (class(d->t) == TKC_STRING)
    {
	/* Store value in existing string, procedure or symbol variable */
	if (d->flag & DT_MULTI && sval(d->t)[0] != '\0')
	{
	    /* MULTISTRING */
	    int ol= strlen((char *)d->t.val);
	    int nl= strlen(val);

	    if (d->t.flag & TKF_FREE)
	    {
		d->t.val= (void *)realloc(sval(d->t),ol+nl+2);
	    }
	    else
	    {
		char *m= (char *)malloc(ol+nl+2);
		strncpy(m, sval(d->t), ol);
		d->t.val= (void *)m;
	    }
	    sval(d->t)[ol]= ',';
	    strcpy(sval(d->t)+ol+1,val);

	    if (copy == SM_FREE) free(val);
	}
	else
	{
	    /* Regular string */
	    if (d->t.flag & TKF_FREE)
		free(sval(d->t));

	    if (copy == SM_COPY)
		d->t.val= (void *)strdup(val);
	    else
		d->t.val= (void *)val;

	    if (copy == SM_KEEP)
		d->t.flag&= ~TKF_FREE;
	    else
		d->t.flag|= TKF_FREE;
	}
    }
    else
	die("query sets variable %s which is not integer, string or time",
	    label);
}


/* FUNC_XDEF:  <value> <label> xdef -
 * Sets <label> to the given value in the current top dictionary.
 */

void func_xdef()
{
Token label, value;

    pop_literal(&label);
    pop_any(&value);
    do_def(&label,&value,0);
}


/* FUNC_DEF:  <label> <value> def -
 * Sets <label> to the given value in the current top dictionary.
 */

void func_def()
{
Token label, value;

    pop_any(&value);
    pop_literal(&label);
    do_def(&label,&value,0);
}

#if 1
/* FUNC_DEFCONSTANT:  <label> <value> defconstant -
 * Sets <label> to the given value in the current top dictionary.
 */

void func_defconstant()
{
Token label,value;

    if (!initializing)
	die("Constants can only be defined from config script");
    pop_any(&value);
    pop_literal(&label);
    do_def(&label,&value,DT_CONSTANT);
}


/* FUNC_XDEFCONSTANT:  <value> <label> xdefconstant -
 * Sets <label> to the given value in the current top dictionary.
 */

void func_xdefconstant()
{
Token label,value;

    if (!initializing)
	die("Constants can only be defined from config script");
    pop_literal(&label);
    pop_any(&value);
    do_def(&label,&value,DT_CONSTANT);
}
#else
void func_defconstant()
{
    die("Panic: defconstant command in compiled program");
}

void func_xdefconstant()
{
    die("Panic: xdefconstant command in compiled program");
}
#endif


void do_def(Token *label, Token *value, int flag)
{
Dict *d;

    if (type(*label) == TK_UNBOUND_LITERAL)
    {
	/* Unbound symbols */
	label_ok(sval(*label));

	if ((d= find_dict(sval(*label),dictstack)) == NULL)
	    (void)add_dict(dictstack,sval(*label),value,flag);
	else
	    set_dict(d,value,flag);
	
	free_val(label);
    }
    else
	/* Bound symbols */
	set_dict(sysdict+ival(*label),value,flag);
}


/* FUNC_XSTORE:  <value> <label> xstore -
 * Sets <label> to the given value in the topmost dictionary where it is
 * defined, or in the current dictionary if it wasn't previously defined.
 */

void func_xstore()
{
Token label, value;

    pop_literal(&label);
    pop_any(&value);
    do_store(&label,&value);
}


/* STORE_DICT - store the given token with as the given variable in the top
 * most dictionary where it is defined, or in the current dictionary if it
 * isn't defined.
 */

void store_dict(char *label, Token *value)
{
Dict *d;

    if ((d= find_dictstack(label)) == NULL)
	(void)add_dict(dictstack,label,value,0);
    else
	set_dict(d,value,0);
}


/* FUNC_STORE:  <label> <value> store -
 * Sets <label> to the given value in the topmost dictionary where it is
 * defined, or in the current dictionary if it wasn't previously defined.
 */

void func_store()
{
Token label, value;

    pop_any(&value);
    pop_literal(&label);
    do_store(&label,&value);
}

void do_store(Token *label, Token *value)
{
    if (type(*label) == TK_UNBOUND_LITERAL)
    {
	/* Unbound symbols */
	label_ok(sval(*label));
	store_dict(sval(*label),value);
	    free_val(label);
    }
    else
	/* Bound symbols */
	set_dict(sysdict+ival(*label),value,0);
}

/* FUNC_UNDEF:  <label> undef -
 * Remove the given label from the topmost dictionary in which it is defined.
 * If it isn't defined, or is only defined in the system dictionary, this is
 * a no-op.
 */

void func_undef()
{
DictStack *s;
Token label;

    if (!pop_literal(&label))
	die("cannot undefine system dictionary symbol %s",
	    sysdict[ival(label)].label);

    for (s= dictstack; s != NULL && s != &sysds; s= s->next)
    {
	if (del_dict(s,sval(label)))
	    return;
    }
}


/* FUNC_INC: <var> inc -
 * Increment an integer varible.   Equivalent to "<var> 1 + <var> xstore",
 * but faster.  It is an error if the variable is undefined or is not an
 * integer.
 */

void func_inc()
{
Token var;
Dict *d= NULL;

    if (!pop_literal(&var))
	/* Bound variable */
	d= sysdict + ival(var);
    else if ((d= find_dictstack(sval(var))) == NULL)
	die("attempt to increment undefined variable %s",sval(var));

    free_val(&var);

    if (type(d->t) != TK_INTEGER)
	die("attempt to increment non-integer variable %s",d->label);

    d->t.val= (void *)(ival(d->t)+1);
}


/* FUNC_DEC: <var> dec -
 * Decrement an integer varible.   Equivalent to "<var> 1 - <var> xstore",
 * but faster.  It is an error if the variable is undefined or is not an
 * integer.
 */

void func_dec()
{
Token var;
Dict *d= NULL;

    if (!pop_literal(&var))
	/* Bound variable */
	d= sysdict + ival(var);
    else if ((d= find_dictstack(sval(var))) == NULL)
	die("attempt to increment undefined variable %s",sval(var));

    free_val(&var);

    if (type(d->t) != TK_INTEGER)
	die("attempt to decrement non-integer variable %s",d->label);

    d->t.val= (void *)(ival(d->t)-1);
}


/* FUNC_DEFINED:  <label> defined <bool>
 * This checks if a label is defined.  It simply pushes true or false on the
 * stack.
 */

void func_defined()
{
Token label;

    /* If label is bound or defined, push a TRUE */
    push_integer(!pop_literal(&label) || get_dict(sval(label)) != NULL);
    free_val(&label);
}


/* FUNC_DEFAULT:  <label> <default-value> default <value>
 * This checks if a label is defined.  If it is, it pushes the value of the
 * label.  If it isn't it pushs the default value.
 */

void func_default()
{
Token deflt, label, *value;

    pop_any(&deflt);
    if (pop_literal(&label))
    {
    	/* Unbound - Is it defined? */
    	if ((value= get_dict(sval(label))) == NULL)
	{
	    /* No, push back the default */
	    push_token(&deflt, FALSE);
	}
	else
	{
	    /* Yes, push the value */
	    free_val(&deflt);
	    push_token(value, TRUE);
	}
    }
    else
    {
    	/* Bound - push its value */
	free_val(&deflt);
	push_token(&sysdict[ival(label)].t, TRUE);
    }
    free_val(&label);
}


/* CSDICT - just like sdict(), but a function.  This is used by modules that
 * are shared with the partutil program.
 */

char *csdict(int i)
{
    return sdict(i);
}


/* FUNC_CONSTANT:  <label> constant <bool>
 * This checks if the given label is a defined constant.  It is always
 * resolved at compile time, so there is no code here.
 */

void func_constant() {}


#ifdef CHECKDICT

/* CHECK_DICT - Do a sanity check on the given dictionary.  This is a debugging
 * tool.
 */

void check_dict(Dict *dict, int size, char *dictname)
{
int i;

    if (dict[0].label[0] == '\0')
    	return;

    for (i= 1; i < size; i++)
    {
    	if (strcmp(dict[i-1].label, dict[i].label) >= 0)
    	    die("sequence problem in %s: %s preceeds %s",
    	    	dictname,dict[i-1].label,dict[i].label);
    }
}


/* CHECK_DICTSTACK - Do a sanity check on all dictionaries in the dictionary
 * stack.  This is a debugging tool.
 */

void check_dictstack()
{
DictStack *ds;
int n;
char bf[BFSZ];

    /* Count number of user dictionaries */
    for (ds= dictstack, n= 0; ds != &sysds && ds != NULL; ds= ds->next, n++)
    {
	sprintf(bf, "dictionary %d", n);
	check_dict(ds->dict, ds->n, bf);
    }

    check_dict(sysdict, SYSDICTLEN, "system dictionary");
}

#endif /* CHECKDICT */
